import os


__all__ = ['PACKAGE_PATH', 'WELCOME', 'AUTHOR', 'PACKAGE_URL']


PACKAGE_PATH = os.path.dirname(__file__)

WELCOME = """
Welcome to **TheKeenan**. Have fun with it!
"""

AUTHOR = """
Author: Bo Zhang (bozhang@nao.cas.cn)
"""

PACKAGE_URL = """
Bitbucket url: https://bitbucket.org/hypergravity/thekeenan  # currently private
Github url:    https://github.com/hypergravity/thekeenan     # in prep.
PYPI url:      https://pypi.python.org/pypi/thekeenan        # in prep.
"""
